/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <tiuser.h>
#include <signal.h>
#include <stropts.h>
#include <wait.h>
#include <fcntl.h>
#include <string.h>

#define tlog(STR) log("%s: %s", (STR), t_errlist[t_errno])
#define MAXACTIVE	1024
#define CMDSZ		256
#define MAXARG		128

struct active {
	pid_t	pid;
	int	fd;
} active[MAXACTIVE];

struct sigaction sa;
sigset_t s;

void doservice(int);
void endconn(int);

extern int t_errno;
extern char *t_errlist[];

extern void log(const char *, ...);

int
getfd(char *transport)
{
	int fd;

	/*
	 * Open the transport provider.
	 */
	if ((fd = t_open(transport, O_RDWR, NULL)) < 0)
		return(-1);

	/*
	 * Bind an arbitrary address to the transport
	 * endpoint.
	 */
	if (t_bind(fd, NULL, NULL) < 0) {
		t_close(fd);
		return(-1);
	}
	return(fd);
}

void
serve(int fd, char *transport)
{
	int afd, i;
	pid_t pid;
	struct t_call *call;

	/*
	 * Allocate a t_call structure to be used when
	 * receiving connect indications.
	 */
	call = (struct t_call *)t_alloc(fd, T_CALL, T_ALL);
	if (call == NULL) {
		tlog("t_alloc failed");
		exit(1);
	}

	/*
	 * Set up SIGCHLD signal handler.
	 */
	sa.sa_handler = endconn;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &sa, NULL);
	sigemptyset(&s);
	sigaddset(&s, SIGCHLD);

	for ( ; ; ) {
		/*
		 * Block until a connect indication arrives.
		 */
		if (t_listen(fd, call) < 0) {
			tlog("t_listen failed");
			exit(1);
		}

		/*
		 * Get a new transport endpoint to be used in
		 * the connection.
		 */
		if ((afd = getfd(transport)) < 0) {
			/*
			 * Can't get a new endpoint -- reject
			 * the connect request.
			 */
			if (t_snddis(fd, call) < 0) {
				tlog("t_snddis failed");
				exit(1);
			}
			continue;
		}

		/*
		 * Accept the connection on afd instead of fd.
		 */
		if (t_accept(fd, afd, call) < 0) {
			if (t_errno == TLOOK) {
				/*
				 * Someone canceled the connect request.
				 */
				if (t_rcvdis(fd, NULL) < 0) {
					tlog("t_rcvdis failed");
					exit(1);
				}
			} else {
				tlog("t_accept failed");
				exit(1);
			}
		} else {
			/*
			 * Critical section: block SIGCHLD.
			 */
			sigprocmask(SIG_BLOCK, &s, NULL);
			if ((pid = fork()) < 0) {
				/*
				 * Fork failed.  Reject the connect request,
				 * unblock SIGCHLD, and continue waiting for
				 * more connect indications.
				 */
				log("fork failed");
				if (t_snddis(afd, NULL) < 0) {
					tlog("t_snddis failed");
					exit(1);
				}
				sigprocmask(SIG_UNBLOCK, &s, NULL);
				continue;
			} else if (pid == 0) {
				/*
				 * The child process handles the service
				 * request and then exits.  First restore
				 * the default disposition of SIGCHLD.
				 */
				sa.sa_handler = SIG_DFL;
				sigemptyset(&sa.sa_mask);
				sa.sa_flags = 0;
				sigaction(SIGCHLD, &sa, NULL);
				sigprocmask(SIG_UNBLOCK, &s, NULL);
				doservice(afd);
			} else {
				/*
				 * The parent remembers the process ID
				 * of the child so it can disconnect the
				 * connection when the child exits.  First
				 * find an unused entry.
				 */
				for (i = 0; i < MAXACTIVE; i++)
					if (active[i].fd == -1)
						break;
				if (i >= MAXACTIVE) {
					log("Too many active children");
					exit(1);
				}
				active[i].pid = pid;
				active[i].fd = afd;

				/*
				 * End of critical section.  Parent unblocks
				 * SIGCHLD and continues to wait for more
				 * connect indications.
				 */
				sigprocmask(SIG_UNBLOCK, &s, NULL);
			}
		}

		/*
		 * After a successful fork, the parent continues
		 * here. Close the file descriptor used in the
		 * connection and continue to listen for more
		 * connect indications.
		 */
		t_close(afd);
	}
}

void
doservice(int fd)
{
	int i, n, flags, nargs;
	char *p, *cmd;
	char buf[CMDSZ];
	char *args[MAXARG];


	/*
	 * Receive the command string.  Must make multiple
	 * calls to t_rcv because the original message
	 * might have been fragmented during transmission.
	 */
	n = t_rcv(fd, buf, CMDSZ, &flags);
	if (n < 0) {
		tlog("Cannot receive request");
		exit(1);
	}

	/*
	 * p is the current location in the command string
	 * and n is the amount remaining to be read.
	 */
	p = &buf[n];
	n = buf[0] - n;
	while (n > 0) {
		i = t_rcv(fd, p, (int)(CMDSZ + p - buf), &flags);
		if (i < 0) {
			tlog("Cannot receive request");
			exit(1);
		}
		p += i;
		n -= i;
	}

	/*
	 * Set up the arguments for exec.
	 */
	cmd = &buf[1];
	p = strrchr(&buf[1], '/');
	if (p)
		args[0] = p+1;
	else
		args[0] = &buf[1];
	nargs = 1;
	n = buf[0] - 1 - strlen(&buf[1]) - 1;
	p = &buf[strlen(&buf[1])+1];
	while (n > 0) {
		args[nargs++] = p;
		n -= strlen(p) + 1;
		p += strlen(p) + 1;
	}

	/*
	 * Prepare to execute the command by making sure
	 * file descriptor 0 is open to /dev/null and file
	 * descriptors 1 and 2 are associated with the
	 * network connection.
	 */
	close(0);
	close(1);
	close(2);
	n = open("/dev/null", O_RDONLY);
	if (n != 0) {
		if (n < 0)
			log("Cannot open /dev/null");
		else
			log("Error setting up stdin");
		exit(0);
	}
	if (dup(fd) != 1) {
		log("Error setting up stdout");
		exit(0);
	}
	if (dup(fd) != 2) {
		log("Error setting up stderr");
		exit(0);
	}

	/*
	 * Push TIRDWR on the stream to get a read/write
	 * interface.
	 */
	if (ioctl(fd, I_PUSH, "tirdwr") < 0) {
		log("Cannot push tirdwr");
		exit(0);
	}

	/*
	 * Close the extra file descriptor open to the
	 * network connection and execute the command.
	 */
	close(fd);
	execvp(cmd, args);
	log("Cannot exec command");
	exit(0);
}

/*
 * SIGCHLD signal handler.
 */
void
endconn(int sig)
{
	int i;
	siginfo_t si;

	while (waitid(P_ALL, 0, &si, WEXITED|WNOHANG) == 0) {
		/*
		 * Find the entry corresponding to the child
		 * process that just terminated.
		 */
		for (i = 0; i < MAXACTIVE; i++) {
			if (active[i].pid == si.si_pid) {
				/*
				 * Found it.  Mark the entry as
				 * unused, pop TIRDWR off the
				 * stream, disconnect the client,
				 * and free the transport endpoint.
				 */
				active[i].pid = 0;
				ioctl(active[i].fd, I_POP, 0);
				t_snddis(active[i].fd, NULL);
				t_close(active[i].fd);
				active[i].fd = -1;
				break;
			}
		}
	}
}
